from MorphologicalAnalysis.FsmMorphologicalAnalyzer cimport FsmMorphologicalAnalyzer
from MorphologicalAnalysis.FsmParse cimport FsmParse
from MorphologicalDisambiguation.RootWordStatistics cimport RootWordStatistics


cdef class AutoDisambiguator:

    cdef FsmMorphologicalAnalyzer morphologicalAnalyzer
    cdef RootWordStatistics rootWordStatistics



from DataStructure.CounterHashMap cimport CounterHashMap
from MorphologicalAnalysis.FsmParseList cimport FsmParseList


cdef class RootWordStatistics:

    cdef dict __statistics

    cpdef bint containsKey(self, str key)
    cpdef CounterHashMap get(self, str key)
    cpdef put(self, str key, CounterHashMap wordStatistics)
    cpdef str bestRootWord(self, FsmParseList parseList, double threshold)
    cpdef saveStatistics(self, str fileName)



from Corpus.Sentence cimport Sentence
from MorphologicalAnalysis.MorphologicalParse cimport MorphologicalParse

from MorphologicalDisambiguation.DisambiguatedWord cimport DisambiguatedWord


cdef class DisambiguationCorpus(Corpus):

    def __init__(self, fileName=None):
        """
        Constructor which creates a list of sentences and a CounterHashMap of wordList.
        """
        cdef list lines
        cdef Sentence newSentence
        cdef str line, word, parse
        cdef DisambiguatedWord newWord
        super().__init__()
        if fileName is not None:
            inputFile = open(fileName, "r", encoding="utf8")
            lines = inputFile.readlines()
            newSentence = Sentence()
            for line in lines:
                word = line[:line.index("\t")]
                parse = line[line.index("\t") + 1:]
                if len(word) > 0 and len(parse) > 0:
                    newWord = DisambiguatedWord(word, MorphologicalParse(parse.strip()))
                    if word == "<S>":
                        newSentence = Sentence()
                    elif word == "</S>":
                        self.addSentence(newSentence)
                    elif word == "<DOC>" or word == "</DOC>" or word == "<TITLE>" or word == "</TITLE>":
                        pass
                    else:
                        newSentence.addWord(newWord)
            inputFile.close()

    cpdef writeToFile(self, str fileName):
        """
        The writeToFile method takes a str file name as an input and writes the elements of sentences list
        to this file with proper tags which indicates the beginnings and endings of the document and sentence.

        PARAMETERS
        ----------
        fileName : str
            File which will be filled with the sentences.
        """
        cdef Sentence sentence
        cdef DisambiguatedWord word
        outputFile = open(fileName, "w", encoding="utf8")
        outputFile.write("<DOC>\t<DOC>+BDTag\n")
        for sentence in self.sentences:
            outputFile.write("<S>\t<S>+BSTag\n")
            for word in sentence.words:
                if isinstance(word, DisambiguatedWord):
                    outputFile.write(word.getName() + "\t" + word.getParse().__str__() + "\n")
            outputFile.write("</S>\t</S>+ESTag\n")
        outputFile.write("</DOC>\t</DOC>+EDTag")
        outputFile.close()



from Dictionary.Word cimport Word
from MorphologicalAnalysis.MorphologicalParse cimport MorphologicalParse


cdef class DisambiguatedWord(Word):

    cdef MorphologicalParse __parse

    cpdef MorphologicalParse getParse(self)



from MorphologicalDisambiguation.AutoDisambiguator import AutoDisambiguator
from MorphologicalDisambiguation.DisambiguationCorpus cimport DisambiguationCorpus
from MorphologicalDisambiguation.MorphologicalDisambiguator cimport MorphologicalDisambiguator
from MorphologicalAnalysis.FsmParseList cimport FsmParseList
from MorphologicalAnalysis.FsmParse cimport FsmParse


cdef class LongestRootFirstDisambiguation(MorphologicalDisambiguator):

    cpdef train(self, DisambiguationCorpus corpus):
        """
        Train method implements method in MorphologicalDisambiguator.

        PARAMETERS
        ----------
        corpus : DisambiguationCorpus
            DisambiguationCorpus to train.
        """
        pass

    cpdef list disambiguate(self, list fsmParses):
        """
        The disambiguate method gets an array of fsmParses. Then loops through that parses and finds the longest root
        word. At the end, gets the parse with longest word among the fsmParses and adds it to the correctFsmParses
        list.

        PARAMETERS
        ----------
        fsmParses : list
            FsmParseList to disambiguate.

        RETURNS
        -------
        list
            CorrectFsmParses list.
        """
        cdef int i
        cdef list correctFsmParses
        cdef FsmParseList fsmParseList
        cdef FsmParse bestParse, newBestParse
        correctFsmParses = []
        i = 0
        for fsmParseList in fsmParses:
            bestParse = fsmParseList.getParseWithLongestRootWord()
            fsmParseList.reduceToParsesWithSameRootAndPos(bestParse.getWordWithPos())
            newBestParse = AutoDisambiguator.caseDisambiguator(i, fsmParses, correctFsmParses)
            if newBestParse is not None:
                bestParse = newBestParse
            correctFsmParses.append(bestParse)
            i = i + 1
        return correctFsmParses

    cpdef saveModel(self):
        pass

    cpdef loadModel(self):
        pass



from MorphologicalDisambiguation.AutoDisambiguator import AutoDisambiguator
from MorphologicalDisambiguation.DisambiguationCorpus cimport DisambiguationCorpus
from MorphologicalDisambiguation.MorphologicalDisambiguator cimport MorphologicalDisambiguator
from MorphologicalAnalysis.FsmParseList cimport FsmParseList
from MorphologicalAnalysis.FsmParse cimport FsmParse
from MorphologicalDisambiguation.RootWordStatistics cimport RootWordStatistics

cdef class RootWordStatisticsDisambiguation(MorphologicalDisambiguator):

    cpdef RootWordStatistics rootWordStatistics

    cpdef train(self, DisambiguationCorpus corpus):
        """
        Train method implements method in MorphologicalDisambiguator.

        PARAMETERS
        ----------
        corpus : DisambiguationCorpus
            DisambiguationCorpus to train.
        """
        self.rootWordStatistics = RootWordStatistics("../penntreebank_statistics.txt")

    cpdef list disambiguate(self, list fsmParses):
        """
        The disambiguate method gets an array of fsmParses. Then loops through that parses and finds the longest root
        word. At the end, gets the parse with longest word among the fsmParses and adds it to the correctFsmParses
        list.

        PARAMETERS
        ----------
        fsmParses : list
            FsmParseList to disambiguate.

        RETURNS
        -------
        list
            CorrectFsmParses list.
        """
        cdef int i
        cdef list correctFsmParses
        cdef FsmParseList fsmParseList
        cdef FsmParse bestParse, newBestParse
        cdef str bestRoot, rootWords
        correctFsmParses = []
        i = 0
        for fsmParseList in fsmParses:
            rootWords = fsmParseList.rootWords()
            if "$" in rootWords:
                bestRoot = self.rootWordStatistics.bestRootWord(fsmParseList, 0.0)
                if bestRoot is None:
                    bestRoot = fsmParseList.getParseWithLongestRootWord().getWord().getName()
            else:
                bestRoot = rootWords
            if bestRoot is not None:
                fsmParseList.reduceToParsesWithSameRoot(bestRoot)
                newBestParse = AutoDisambiguator.caseDisambiguator(i, fsmParses, correctFsmParses)
                if newBestParse is not None:
                    bestParse = newBestParse
                else:
                    bestParse = fsmParseList.getFsmParse(0)
            else:
                bestParse = fsmParseList.getFsmParse(0)
            correctFsmParses.append(bestParse)
            i = i + 1
        return correctFsmParses

    cpdef saveModel(self):
        pass

    cpdef loadModel(self):
        pass



cdef class RootWordStatistics:

    def __init__(self, fileName=None):
        """
        Constructor of RootWordStatistics class that generates a new map for statistics.
        """
        cdef int i, count, j
        cdef str line, rootWord
        cdef list items
        cdef CounterHashMap wordMap
        self.__statistics = {}
        if fileName is not None:
            inputFile = open(fileName, encoding="utf8")
            size = int(inputFile.readline().strip())
            for i in range(size):
                line = inputFile.readline().strip()
                items = line.split()
                rootWord = items[0]
                count = int(items[1])
                wordMap = CounterHashMap()
                for j in range(count):
                    line = inputFile.readline().strip()
                    items = line.split()
                    wordMap.putNTimes(items[0], int(items[1]))
                self.__statistics[rootWord] = wordMap
            inputFile.close()

    cpdef bint containsKey(self, str key):
        """
        Method to check whether statistics contains the given String.

        PARAMETERS
        ----------
        key : str
            String to look for.

        RETURNS
        -------
        bool
            Returns True if this map contains a mapping for the specified key.
        """
        return key in self.__statistics

    cpdef CounterHashMap get(self, str key):
        """
        Method to get the value of the given String.

        PARAMETERS
        ----------
        key : str
            String to look for.

        RETURNS
        -------
        CounterHashMap
            Returns the value to which the specified key is mapped, or None if this map contains no mapping for the key.
        """
        return self.__statistics[key]

    cpdef put(self, str key, CounterHashMap wordStatistics):
        """
        Method to associates a String along with a CounterHashMap in the statistics.

        PARAMETERS
        ----------
        key : str
            Key with which the specified value is to be associated.
        wordStatistics : CounterHashMap
            Value to be associated with the specified key.
        """
        self.__statistics[key] = wordStatistics

    cpdef str bestRootWord(self, FsmParseList parseList, double threshold):
        """
        The bestRootWord method gets the root word of given FsmParseList and if statistics has a value for that word,
        it returns the max value associated with that word.

        PARAMETERS
        ----------
        parseList : FsmParseList
            FsmParseList to check.
        threshold : float
            A double value for limit.

        RETURNS
        -------
        str
            The max value for the root word.
        """
        cdef int i
        cdef str surfaceForm, rootWord
        cdef CounterHashMap rootWordStatistics
        surfaceForm = parseList.getFsmParse(0).getSurfaceForm()
        if surfaceForm in self.__statistics:
            rootWordStatistics = self.__statistics[surfaceForm]
            rootWord = rootWordStatistics.max(threshold)
            for i in range(parseList.size()):
                if parseList.getFsmParse(i).getWord().getName() == rootWord:
                    return rootWord
        return None

    cpdef saveStatistics(self, str fileName):
        """
        Method to save statistics into file specified with the input file name.

        PARAMETERS
        ----------
        fileName : str
            File to save the statistics.
        """
        cdef str rootWord
        cdef CounterHashMap wordMap
        outputFile = open(fileName, mode="w", encoding="utf8")
        outputFile.write(len(self.__statistics).__str__() + "\n")
        for rootWord in self.__statistics:
            wordMap = self.__statistics[rootWord]
            outputFile.write(rootWord + " " + len(wordMap).__str__() + "\n")
            outputFile.write(wordMap.__str__())
        outputFile.close()



from distutils.core import setup
from Cython.Build import cythonize

setup(
    ext_modules=cythonize(["MorphologicalDisambiguation/*.pyx"],
                          compiler_directives={'language_level': "3"}),
    name='NlpToolkit-MorphologicalDisambiguation-Cy',
    version='1.0.3',
    packages=['MorphologicalDisambiguation'],
    package_data={'MorphologicalDisambiguation': ['*.pxd', '*.pyx', '*.c']},
    url='https://github.com/olcaytaner/TurkishMorphologicalDisambiguation-Cy',
    license='',
    author='olcay',
    author_email='olcaytaner@isikun.edu.tr',
    description='Turkish Morphological Disambiguation Library',
    install_requires=['NlpToolkit-MorphologicalAnalysis-Cy', 'NlpToolkit-NGram-Cy']
)
